# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import concurrent.futures
import logging
import os
import typing
from json import dumps

LabelValue = typing.Union[str, bool, int, float]
Labels = typing.Dict[str, LabelValue]
logger = logging.getLogger(__name__)


class Resource:
    def __init__(self, labels: Labels):
        self._labels = labels.copy()

    @staticmethod
    def create(labels: Labels) -> "Resource":
        if not labels:
            return _EMPTY_RESOURCE
        return Resource(labels)

    @staticmethod
    def create_empty() -> "Resource":
        return _EMPTY_RESOURCE

    @property
    def labels(self) -> Labels:
        return self._labels.copy()

    def merge(self, other: "Resource") -> "Resource":
        merged_labels = self.labels
        # pylint: disable=protected-access
        for key, value in other._labels.items():
            if key not in merged_labels or merged_labels[key] == "":
                merged_labels[key] = value
        return Resource(merged_labels)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Resource):
            return False
        return self._labels == other._labels

    def __hash__(self):
        return hash(dumps(self._labels, sort_keys=True))


_EMPTY_RESOURCE = Resource({})


class ResourceDetector(abc.ABC):
    def __init__(self, raise_on_error=False):
        self.raise_on_error = raise_on_error

    @abc.abstractmethod
    def detect(self) -> "Resource":
        raise NotImplementedError()


class OTELResourceDetector(ResourceDetector):
    # pylint: disable=no-self-use
    def detect(self) -> "Resource":
        env_resources_items = os.environ.get("OTEL_RESOURCE_ATTRIBUTES")
        env_resource_map = {}
        if env_resources_items:
            env_resource_map = {
                key.strip(): value.strip()
                for key, value in (
                    item.split("=") for item in env_resources_items.split(",")
                )
            }
        return Resource(env_resource_map)


def get_aggregated_resources(
    detectors: typing.List["ResourceDetector"],
    initial_resource: typing.Optional[Resource] = None,
    timeout=5,
) -> "Resource":
    """ Retrieves resources from detectors in the order that they were passed

    :param detectors: List of resources in order of priority
    :param initial_resource: Static resource. This has highest priority
    :param timeout: Number of seconds to wait for each detector to return
    :return:
    """
    final_resource = initial_resource or _EMPTY_RESOURCE
    detectors = [OTELResourceDetector()] + detectors

    with concurrent.futures.ThreadPoolExecutor(max_workers=4) as executor:
        futures = [executor.submit(detector.detect) for detector in detectors]
        for detector_ind, future in enumerate(futures):
            detector = detectors[detector_ind]
            try:
                detected_resources = future.result(timeout=timeout)
            # pylint: disable=broad-except
            except Exception as ex:
                if detector.raise_on_error:
                    raise ex
                logger.warning(
                    "Exception %s in detector %s, ignoring", ex, detector
                )
                detected_resources = _EMPTY_RESOURCE
            finally:
                final_resource = final_resource.merge(detected_resources)
    return final_resource



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Sequence

from opentelemetry.sdk.metrics.export import MetricRecord
from opentelemetry.sdk.util import get_dict_as_key


class Batcher:
    """Base class for all batcher types.

    The batcher is responsible for storing the aggregators and aggregated
    values received from updates from metrics in the meter. The stored values
    will be sent to an exporter for exporting.
    """

    def __init__(self, stateful: bool):
        self._batch_map = {}
        # stateful=True indicates the batcher computes checkpoints from over
        # the process lifetime. False indicates the batcher computes
        # checkpoints which describe the updates of a single collection period
        # (deltas)
        self.stateful = stateful

    def checkpoint_set(self) -> Sequence[MetricRecord]:
        """Returns a list of MetricRecords used for exporting.

        The list of MetricRecords is a snapshot created from the current
        data in all of the aggregators in this batcher.
        """
        metric_records = []
        # pylint: disable=W0612
        for (
            (instrument, aggregator_type, _, labels),
            aggregator,
        ) in self._batch_map.items():
            metric_records.append(MetricRecord(instrument, labels, aggregator))
        return metric_records

    def finished_collection(self):
        """Performs certain post-export logic.

        For batchers that are stateless, resets the batch map.
        """
        if not self.stateful:
            self._batch_map = {}

    def process(self, record) -> None:
        """Stores record information to be ready for exporting."""
        # Checkpoints the current aggregator value to be collected for export
        aggregator = record.aggregator
        aggregator.take_checkpoint()

        # The uniqueness of a batch record is defined by a specific metric
        # using an aggregator type with a specific set of labels.
        # If two aggregators are the same but with different configs, they are still two valid unique records
        # (for example, two histogram views with different buckets)
        key = (
            record.instrument,
            aggregator.__class__,
            get_dict_as_key(aggregator.config),
            record.labels,
        )

        batch_value = self._batch_map.get(key)

        if batch_value:
            # Update the stored checkpointed value if exists. The call to merge
            # here combines only identical records (same key).
            batch_value.merge(aggregator)
            return

        # create a copy of the aggregator and update
        # it with the current checkpointed value for long-term storage
        aggregator = record.aggregator.__class__(
            config=record.aggregator.config
        )
        aggregator.merge(record.aggregator)

        self._batch_map[key] = aggregator



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Sequence, Tuple

from opentelemetry import metrics as metrics_api
from opentelemetry.sdk.metrics.export.aggregate import Aggregator


class MetricsExportResult(Enum):
    SUCCESS = 0
    FAILURE = 1


class MetricRecord:
    def __init__(
        self,
        instrument: metrics_api.InstrumentT,
        labels: Tuple[Tuple[str, str]],
        aggregator: Aggregator,
    ):
        self.instrument = instrument
        self.labels = labels
        self.aggregator = aggregator


class MetricsExporter:
    """Interface for exporting metrics.

    Interface to be implemented by services that want to export recorded
    metrics in its own format.
    """

    def export(
        self, metric_records: Sequence[MetricRecord]
    ) -> "MetricsExportResult":
        """Exports a batch of telemetry data.

        Args:
            metric_records: A sequence of `MetricRecord` s. A `MetricRecord`
                contains the metric to be exported, the labels associated
                with that metric, as well as the aggregator used to export the
                current checkpointed value.

        Returns:
            The result of the export
        """

    def shutdown(self) -> None:
        """Shuts down the exporter.

        Called when the SDK is shut down.
        """


class ConsoleMetricsExporter(MetricsExporter):
    """Implementation of `MetricsExporter` that prints metrics to the console.

    This class can be used for diagnostic purposes. It prints the exported
    metrics to the console STDOUT.
    """

    def export(
        self, metric_records: Sequence[MetricRecord]
    ) -> "MetricsExportResult":
        for record in metric_records:
            print(
                '{}(data="{}", labels="{}", value={})'.format(
                    type(self).__name__,
                    record.instrument,
                    record.labels,
                    record.aggregator.checkpoint,
                )
            )
        return MetricsExportResult.SUCCESS



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import logging
import threading
from collections import OrderedDict, namedtuple

from opentelemetry.util import time_ns

logger = logging.getLogger(__name__)


class Aggregator(abc.ABC):
    """Base class for aggregators.

    Aggregators are responsible for holding aggregated values and taking a
    snapshot of these values upon export (checkpoint).
    """

    def __init__(self, config=None):
        self.current = None
        self.checkpoint = None
        if config:
            self.config = config
        else:
            self.config = {}

    @abc.abstractmethod
    def update(self, value):
        """Updates the current with the new value."""

    @abc.abstractmethod
    def take_checkpoint(self):
        """Stores a snapshot of the current value."""

    @abc.abstractmethod
    def merge(self, other):
        """Combines two aggregator values."""


class SumAggregator(Aggregator):
    """Aggregator for counter metrics."""

    def __init__(self, config=None):
        super().__init__(config=config)
        self.current = 0
        self.checkpoint = 0
        self._lock = threading.Lock()
        self.last_update_timestamp = None

    def update(self, value):
        with self._lock:
            self.current += value
            self.last_update_timestamp = time_ns()

    def take_checkpoint(self):
        with self._lock:
            self.checkpoint = self.current
            self.current = 0

    def merge(self, other):
        if verify_type(self, other):
            with self._lock:
                self.checkpoint += other.checkpoint
                self.last_update_timestamp = get_latest_timestamp(
                    self.last_update_timestamp, other.last_update_timestamp
                )


class MinMaxSumCountAggregator(Aggregator):
    """Aggregator for ValueRecorder metrics that keeps min, max, sum, count."""

    _TYPE = namedtuple("minmaxsumcount", "min max sum count")
    _EMPTY = _TYPE(None, None, None, 0)

    @classmethod
    def _merge_checkpoint(cls, val1, val2):
        if val1 is cls._EMPTY:
            return val2
        if val2 is cls._EMPTY:
            return val1
        return cls._TYPE(
            min(val1.min, val2.min),
            max(val1.max, val2.max),
            val1.sum + val2.sum,
            val1.count + val2.count,
        )

    def __init__(self, config=None):
        super().__init__(config=config)
        self.current = self._EMPTY
        self.checkpoint = self._EMPTY
        self._lock = threading.Lock()
        self.last_update_timestamp = None

    def update(self, value):
        with self._lock:
            if self.current is self._EMPTY:
                self.current = self._TYPE(value, value, value, 1)
            else:
                self.current = self._TYPE(
                    min(self.current.min, value),
                    max(self.current.max, value),
                    self.current.sum + value,
                    self.current.count + 1,
                )
            self.last_update_timestamp = time_ns()

    def take_checkpoint(self):
        with self._lock:
            self.checkpoint = self.current
            self.current = self._EMPTY

    def merge(self, other):
        if verify_type(self, other):
            with self._lock:
                self.checkpoint = self._merge_checkpoint(
                    self.checkpoint, other.checkpoint
                )
                self.last_update_timestamp = get_latest_timestamp(
                    self.last_update_timestamp, other.last_update_timestamp
                )


class HistogramAggregator(Aggregator):
    """Agregator for ValueRecorder metrics that keeps a histogram of values."""

    def __init__(self, config=None):
        super().__init__(config=config)
        self._lock = threading.Lock()
        self.last_update_timestamp = None
        boundaries = self.config.get("bounds")
        if boundaries and self._validate_boundaries(boundaries):
            self._boundaries = boundaries
        else:
            # no buckets except < 0 and >
            self._boundaries = (0,)

        self.current = OrderedDict([(bb, 0) for bb in self._boundaries])
        self.checkpoint = OrderedDict([(bb, 0) for bb in self._boundaries])

        self.current[">"] = 0
        self.checkpoint[">"] = 0

    # pylint: disable=R0201
    def _validate_boundaries(self, boundaries):
        if not boundaries:
            logger.warning("Bounds is empty. Using default.")
            return False
        if not all(
            boundaries[ii] < boundaries[ii + 1]
            for ii in range(len(boundaries) - 1)
        ):
            logger.warning(
                "Bounds must be sorted in increasing order. Using default."
            )
            return False
        return True

    @classmethod
    def _merge_checkpoint(cls, val1, val2):
        if val1.keys() == val2.keys():
            for ii, bb in val2.items():
                val1[ii] += bb
        else:
            logger.warning("Cannot merge histograms with different buckets.")
        return val1

    def update(self, value):
        with self._lock:
            if self.current is None:
                self.current = [0 for ii in range(len(self._boundaries) + 1)]
            # greater than max value
            if value >= self._boundaries[len(self._boundaries) - 1]:
                self.current[">"] += 1
            else:
                for bb in self._boundaries:
                    # find first bucket that value is less than
                    if value < bb:
                        self.current[bb] += 1
                        break
            self.last_update_timestamp = time_ns()

    def take_checkpoint(self):
        with self._lock:
            self.checkpoint = self.current
            self.current = OrderedDict([(bb, 0) for bb in self._boundaries])
            self.current[">"] = 0

    def merge(self, other):
        if verify_type(self, other):
            with self._lock:
                self.checkpoint = self._merge_checkpoint(
                    self.checkpoint, other.checkpoint
                )
                self.last_update_timestamp = get_latest_timestamp(
                    self.last_update_timestamp, other.last_update_timestamp
                )


class LastValueAggregator(Aggregator):
    """Aggregator that stores last value results."""

    def __init__(self, config=None):
        super().__init__(config=config)
        self._lock = threading.Lock()
        self.last_update_timestamp = None

    def update(self, value):
        with self._lock:
            self.current = value
            self.last_update_timestamp = time_ns()

    def take_checkpoint(self):
        with self._lock:
            self.checkpoint = self.current
            self.current = None

    def merge(self, other):
        last = self.checkpoint
        self.last_update_timestamp = get_latest_timestamp(
            self.last_update_timestamp, other.last_update_timestamp
        )
        if self.last_update_timestamp == other.last_update_timestamp:
            last = other.checkpoint
        self.checkpoint = last


class ValueObserverAggregator(Aggregator):
    """Same as MinMaxSumCount but also with last value."""

    _TYPE = namedtuple("minmaxsumcountlast", "min max sum count last")

    def __init__(self, config=None):
        super().__init__(config=config)
        self.mmsc = MinMaxSumCountAggregator()
        self.current = None
        self.checkpoint = self._TYPE(None, None, None, 0, None)
        self.last_update_timestamp = None

    def update(self, value):
        self.mmsc.update(value)
        self.current = value
        self.last_update_timestamp = time_ns()

    def take_checkpoint(self):
        self.mmsc.take_checkpoint()
        self.checkpoint = self._TYPE(*(self.mmsc.checkpoint + (self.current,)))

    def merge(self, other):
        if verify_type(self, other):
            self.mmsc.merge(other.mmsc)
            last = self.checkpoint.last
            self.last_update_timestamp = get_latest_timestamp(
                self.last_update_timestamp, other.last_update_timestamp
            )
            if self.last_update_timestamp == other.last_update_timestamp:
                last = other.checkpoint.last
            self.checkpoint = self._TYPE(*(self.mmsc.checkpoint + (last,)))


def get_latest_timestamp(time_stamp, other_timestamp):
    if time_stamp is None:
        return other_timestamp
    if other_timestamp is not None:
        if time_stamp < other_timestamp:
            return other_timestamp
    return time_stamp


# pylint: disable=R1705
def verify_type(this, other):
    if isinstance(other, this.__class__):
        return True
    else:
        logger.warning(
            "Error in merging %s with %s.",
            this.__class__.__name__,
            other.__class__.__name__,
        )
        return False



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from opentelemetry.sdk import metrics
from opentelemetry.sdk.metrics import Counter, ValueRecorder, view
from opentelemetry.sdk.metrics.export import aggregate
from opentelemetry.sdk.metrics.export.aggregate import (
    HistogramAggregator,
    MinMaxSumCountAggregator,
    SumAggregator,
)
from opentelemetry.sdk.metrics.export.controller import PushController
from opentelemetry.sdk.metrics.export.in_memory_metrics_exporter import (
    InMemoryMetricsExporter,
)
from opentelemetry.sdk.metrics.view import View, ViewConfig


class TestUtil(unittest.TestCase):
    @mock.patch("opentelemetry.sdk.metrics.view.logger")
    def test_default_aggregator(self, logger_mock):
        meter = metrics.MeterProvider().get_meter(__name__)
        counter = metrics.Counter("", "", "1", int, meter)
        self.assertEqual(
            view.get_default_aggregator(counter), aggregate.SumAggregator
        )
        ud_counter = metrics.UpDownCounter("", "", "1", int, meter)
        self.assertEqual(
            view.get_default_aggregator(ud_counter), aggregate.SumAggregator
        )
        observer = metrics.SumObserver(lambda: None, "", "", "1", int)
        self.assertEqual(
            view.get_default_aggregator(observer),
            aggregate.LastValueAggregator,
        )
        ud_observer = metrics.SumObserver(lambda: None, "", "", "1", int)
        self.assertEqual(
            view.get_default_aggregator(ud_observer),
            aggregate.LastValueAggregator,
        )
        recorder = metrics.ValueRecorder("", "", "1", int, meter)
        self.assertEqual(
            view.get_default_aggregator(recorder),
            aggregate.MinMaxSumCountAggregator,
        )
        v_observer = metrics.ValueObserver(lambda: None, "", "", "1", int)
        self.assertEqual(
            view.get_default_aggregator(v_observer),
            aggregate.ValueObserverAggregator,
        )
        self.assertEqual(
            view.get_default_aggregator(DummyMetric()), aggregate.SumAggregator
        )
        self.assertEqual(logger_mock.warning.call_count, 1)


class TestStateless(unittest.TestCase):
    def setUp(self):
        self.meter = metrics.MeterProvider(stateful=False).get_meter(__name__)
        self.exporter = InMemoryMetricsExporter()
        self.controller = PushController(self.meter, self.exporter, 30)

    def tearDown(self):
        self.controller.shutdown()

    def test_label_keys(self):
        test_counter = self.meter.create_metric(
            name="test_counter",
            description="description",
            unit="By",
            value_type=int,
            metric_type=Counter,
        )
        counter_view = View(
            test_counter,
            SumAggregator,
            label_keys=["environment"],
            view_config=ViewConfig.LABEL_KEYS,
        )

        self.meter.register_view(counter_view)
        test_counter.add(6, {"environment": "production", "customer_id": 123})
        test_counter.add(5, {"environment": "production", "customer_id": 247})

        self.controller.tick()

        metric_data = self.exporter.get_exported_metrics()
        self.assertEqual(len(metric_data), 1)
        self.assertEqual(
            metric_data[0].labels, (("environment", "production"),)
        )
        self.assertEqual(metric_data[0].aggregator.checkpoint, 11)

    def test_ungrouped(self):
        test_counter = self.meter.create_metric(
            name="test_counter",
            description="description",
            unit="By",
            value_type=int,
            metric_type=Counter,
        )
        counter_view = View(
            test_counter,
            SumAggregator,
            label_keys=["environment"],
            view_config=ViewConfig.UNGROUPED,
        )

        self.meter.register_view(counter_view)
        test_counter.add(6, {"environment": "production", "customer_id": 123})
        test_counter.add(5, {"environment": "production", "customer_id": 247})

        self.controller.tick()

        metric_data = self.exporter.get_exported_metrics()
        data_set = set()
        for data in metric_data:
            data_set.add((data.labels, data.aggregator.checkpoint))
        self.assertEqual(len(metric_data), 2)
        label1 = (("customer_id", 123), ("environment", "production"))
        label2 = (("customer_id", 247), ("environment", "production"))
        self.assertTrue((label1, 6) in data_set)
        self.assertTrue((label2, 5) in data_set)

    def test_multiple_views(self):
        test_counter = self.meter.create_metric(
            name="test_counter",
            description="description",
            unit="By",
            value_type=int,
            metric_type=Counter,
        )

        counter_view = View(
            test_counter,
            SumAggregator,
            label_keys=["environment"],
            view_config=ViewConfig.UNGROUPED,
        )

        mmsc_view = View(
            test_counter,
            MinMaxSumCountAggregator,
            label_keys=["environment"],
            view_config=ViewConfig.LABEL_KEYS,
        )

        self.meter.register_view(counter_view)
        self.meter.register_view(mmsc_view)
        test_counter.add(6, {"environment": "production", "customer_id": 123})
        test_counter.add(5, {"environment": "production", "customer_id": 247})

        self.controller.tick()

        metric_data = self.exporter.get_exported_metrics()
        sum_set = set()
        mmsc_set = set()
        for data in metric_data:
            if isinstance(data.aggregator, SumAggregator):
                tup = (data.labels, data.aggregator.checkpoint)
                sum_set.add(tup)
            elif isinstance(data.aggregator, MinMaxSumCountAggregator):
                mmsc_set.add(data)
                self.assertEqual(data.labels, (("environment", "production"),))
                self.assertEqual(data.aggregator.checkpoint.sum, 11)
        # we have to assert this way because order is unknown
        self.assertEqual(len(sum_set), 2)
        self.assertEqual(len(mmsc_set), 1)
        label1 = (("customer_id", 123), ("environment", "production"))
        label2 = (("customer_id", 247), ("environment", "production"))
        self.assertTrue((label1, 6) in sum_set)
        self.assertTrue((label2, 5) in sum_set)


class TestHistogramView(unittest.TestCase):
    def test_histogram_stateful(self):
        meter = metrics.MeterProvider(stateful=True).get_meter(__name__)
        exporter = InMemoryMetricsExporter()
        controller = PushController(meter, exporter, 30)

        requests_size = meter.create_metric(
            name="requests_size",
            description="size of requests",
            unit="1",
            value_type=int,
            metric_type=ValueRecorder,
        )

        size_view = View(
            requests_size,
            HistogramAggregator,
            aggregator_config={"bounds": [20, 40, 60, 80, 100]},
            label_keys=["environment"],
            view_config=ViewConfig.LABEL_KEYS,
        )

        meter.register_view(size_view)

        # Since this is using the HistogramAggregator, the bucket counts will be reflected
        # with each record
        requests_size.record(25, {"environment": "staging", "test": "value"})
        requests_size.record(1, {"environment": "staging", "test": "value2"})
        requests_size.record(200, {"environment": "staging", "test": "value3"})

        controller.tick()

        metrics_list = exporter.get_exported_metrics()
        self.assertEqual(len(metrics_list), 1)
        checkpoint = metrics_list[0].aggregator.checkpoint
        self.assertEqual(
            tuple(checkpoint.items()),
            ((20, 1), (40, 1), (60, 0), (80, 0), (100, 0), (">", 1)),
        )
        exporter.clear()

        requests_size.record(25, {"environment": "staging", "test": "value"})
        requests_size.record(1, {"environment": "staging", "test": "value2"})
        requests_size.record(200, {"environment": "staging", "test": "value3"})

        controller.tick()

        metrics_list = exporter.get_exported_metrics()
        self.assertEqual(len(metrics_list), 1)
        checkpoint = metrics_list[0].aggregator.checkpoint
        self.assertEqual(
            tuple(checkpoint.items()),
            ((20, 2), (40, 2), (60, 0), (80, 0), (100, 0), (">", 2)),
        )

    def test_histogram_stateless(self):
        # Use the meter type provided by the SDK package
        meter = metrics.MeterProvider(stateful=False).get_meter(__name__)
        exporter = InMemoryMetricsExporter()
        controller = PushController(meter, exporter, 30)

        requests_size = meter.create_metric(
            name="requests_size",
            description="size of requests",
            unit="1",
            value_type=int,
            metric_type=ValueRecorder,
        )

        size_view = View(
            requests_size,
            HistogramAggregator,
            aggregator_config={"bounds": [20, 40, 60, 80, 100]},
            label_keys=["environment"],
            view_config=ViewConfig.LABEL_KEYS,
        )

        meter.register_view(size_view)

        # Since this is using the HistogramAggregator, the bucket counts will be reflected
        # with each record
        requests_size.record(25, {"environment": "staging", "test": "value"})
        requests_size.record(1, {"environment": "staging", "test": "value2"})
        requests_size.record(200, {"environment": "staging", "test": "value3"})

        controller.tick()

        metrics_list = exporter.get_exported_metrics()
        self.assertEqual(len(metrics_list), 1)
        checkpoint = metrics_list[0].aggregator.checkpoint
        self.assertEqual(
            tuple(checkpoint.items()),
            ((20, 1), (40, 1), (60, 0), (80, 0), (100, 0), (">", 1)),
        )
        exporter.clear()

        requests_size.record(25, {"environment": "staging", "test": "value"})
        requests_size.record(1, {"environment": "staging", "test": "value2"})
        requests_size.record(200, {"environment": "staging", "test": "value3"})

        controller.tick()

        metrics_list = exporter.get_exported_metrics()
        self.assertEqual(len(metrics_list), 1)
        checkpoint = metrics_list[0].aggregator.checkpoint
        self.assertEqual(
            tuple(checkpoint.items()),
            ((20, 1), (40, 1), (60, 0), (80, 0), (100, 0), (">", 1)),
        )


class DummyMetric(metrics.Metric):
    # pylint: disable=W0231
    def __init__(self):
        pass



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
from collections import defaultdict
from typing import Optional, Sequence, Tuple

from opentelemetry.metrics import (
    Counter,
    InstrumentT,
    SumObserver,
    UpDownCounter,
    UpDownSumObserver,
    ValueObserver,
    ValueRecorder,
    ValueT,
)
from opentelemetry.sdk.metrics.export.aggregate import (
    Aggregator,
    LastValueAggregator,
    MinMaxSumCountAggregator,
    SumAggregator,
    ValueObserverAggregator,
)

logger = logging.getLogger(__name__)


class ViewData:
    def __init__(self, labels: Tuple[Tuple[str, str]], aggregator: Aggregator):
        self.labels = labels
        self.aggregator = aggregator

    def record(self, value: ValueT):
        self.aggregator.update(value)

    # Uniqueness is based on labels and aggregator type
    def __hash__(self):
        return hash((self.labels, self.aggregator.__class__))

    def __eq__(self, other):
        return (
            self.labels == other.labels
            and self.aggregator.__class__ == other.aggregator.__class__
        )


class ViewConfig:

    UNGROUPED = 0
    LABEL_KEYS = 1
    DROP_ALL = 2


class View:
    def __init__(
        self,
        metric: InstrumentT,
        aggregator: type,
        aggregator_config: Optional[dict] = None,
        label_keys: Optional[Sequence[str]] = None,
        view_config: ViewConfig = ViewConfig.UNGROUPED,
    ):
        self.metric = metric
        self.aggregator = aggregator
        if aggregator_config is None:
            aggregator_config = {}
        self.aggregator_config = aggregator_config
        if label_keys is None:
            label_keys = []
        self.label_keys = sorted(label_keys)
        self.view_config = view_config
        self.view_datas = set()

    def get_view_data(self, labels):
        """Find an existing ViewData for this set of labels. If that ViewData
            does not exist, create a new one to represent the labels
        """
        active_labels = []
        if self.view_config == ViewConfig.LABEL_KEYS:
            # reduce the set of labels to only labels specified in label_keys
            active_labels = {
                (lk, lv) for lk, lv in labels if lk in set(self.label_keys)
            }
            active_labels = tuple(active_labels)
        elif self.view_config == ViewConfig.UNGROUPED:
            active_labels = labels

        for view_data in self.view_datas:
            if view_data.labels == active_labels:
                return view_data
        new_view_data = ViewData(
            active_labels, self.aggregator(self.aggregator_config)
        )
        self.view_datas.add(new_view_data)
        return new_view_data

    # Uniqueness is based on metric, aggregator type, aggregator config,
    # ordered label keys and ViewConfig
    def __hash__(self):
        return hash(
            (
                self.metric,
                self.aggregator.__class__,
                tuple(self.label_keys),
                tuple(self.aggregator_config),
                self.view_config,
            )
        )

    def __eq__(self, other):
        return (
            self.metric == other.metric
            and self.aggregator.__class__ == other.aggregator.__class__
            and self.label_keys == other.label_keys
            and self.aggregator_config == other.aggregator_config
            and self.view_config == other.view_config
        )


class ViewManager:
    def __init__(self):
        self.views = defaultdict(set)  # Map[Metric, Set]
        self._view_lock = threading.Lock()
        self.view_datas = set()

    def register_view(self, view):
        with self._view_lock:
            if view not in self.views[view.metric]:
                self.views[view.metric].add(view)
            else:
                logger.warning("View already registered.")
                return

    def unregister_view(self, view):
        with self._view_lock:
            if self.views.get(view.metric) is None:
                logger.warning("Metric for view does not exist.")
            elif view in self.views.get(view.metric):
                self.views.get(view.metric).remove(view)

    def get_view_datas(self, metric, labels):
        view_datas = set()
        views = self.views.get(metric)
        # No views configured, use default aggregations
        if views is None:
            # make a default view for the metric
            default_view = View(metric, get_default_aggregator(metric))
            self.register_view(default_view)
            views = [default_view]

        for view in views:
            view_datas.add(view.get_view_data(labels))

        return view_datas


def get_default_aggregator(instrument: InstrumentT) -> Aggregator:
    """Returns an aggregator based on metric instrument's type.

    Aggregators keep track of and updates values when metrics get updated.
    """
    # pylint:disable=R0201
    instrument_type = instrument.__class__
    if issubclass(instrument_type, (Counter, UpDownCounter)):
        return SumAggregator
    if issubclass(instrument_type, (SumObserver, UpDownSumObserver)):
        return LastValueAggregator
    if issubclass(instrument_type, ValueRecorder):
        return MinMaxSumCountAggregator
    if issubclass(instrument_type, ValueObserver):
        return ValueObserverAggregator
    logger.warning("No default aggregator configured for: %s", instrument_type)
    return SumAggregator



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unittest import mock

from opentelemetry import metrics as metrics_api
from opentelemetry.sdk import metrics, resources
from opentelemetry.sdk.metrics.export.aggregate import (
    MinMaxSumCountAggregator,
    SumAggregator,
)
from opentelemetry.sdk.metrics.view import View


class TestMeterProvider(unittest.TestCase):
    def test_stateful(self):
        meter_provider = metrics.MeterProvider(stateful=False)
        meter = meter_provider.get_meter(__name__)
        self.assertIs(meter.batcher.stateful, False)

    def test_resource(self):
        resource = resources.Resource.create({})
        meter_provider = metrics.MeterProvider(resource=resource)
        meter = meter_provider.get_meter(__name__)
        self.assertIs(meter.resource, resource)

    def test_resource_empty(self):
        meter_provider = metrics.MeterProvider()
        meter = meter_provider.get_meter(__name__)
        # pylint: disable=protected-access
        self.assertIs(meter.resource, resources._EMPTY_RESOURCE)

    def test_start_pipeline(self):
        exporter = mock.Mock()
        meter_provider = metrics.MeterProvider()
        meter = meter_provider.get_meter(__name__)
        # pylint: disable=protected-access
        meter_provider.start_pipeline(meter, exporter, 6)
        try:
            self.assertEqual(len(meter_provider._exporters), 1)
            self.assertEqual(len(meter_provider._controllers), 1)
        finally:
            meter_provider.shutdown()

    def test_shutdown(self):
        controller = mock.Mock()
        exporter = mock.Mock()
        meter_provider = metrics.MeterProvider()
        # pylint: disable=protected-access
        meter_provider._controllers = [controller]
        meter_provider._exporters = [exporter]
        meter_provider.shutdown()
        self.assertEqual(controller.shutdown.call_count, 1)
        self.assertEqual(exporter.shutdown.call_count, 1)
        self.assertIsNone(meter_provider._atexit_handler)


class TestMeter(unittest.TestCase):
    def test_extends_api(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        self.assertIsInstance(meter, metrics_api.Meter)

    def test_collect_metrics(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        batcher_mock = mock.Mock()
        meter.batcher = batcher_mock
        counter = meter.create_metric(
            "name", "desc", "unit", float, metrics.Counter
        )
        labels = {"key1": "value1"}
        meter.register_view(View(counter, SumAggregator))
        counter.add(1.0, labels)
        meter.collect()
        self.assertTrue(batcher_mock.process.called)

    def test_collect_no_metrics(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        batcher_mock = mock.Mock()
        meter.batcher = batcher_mock
        meter.collect()
        self.assertFalse(batcher_mock.process.called)

    def test_collect_not_registered(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        batcher_mock = mock.Mock()
        meter.batcher = batcher_mock
        counter = metrics.Counter("name", "desc", "unit", float, meter)
        labels = {"key1": "value1"}
        counter.add(1.0, labels)
        meter.collect()
        self.assertFalse(batcher_mock.process.called)

    def test_collect_disabled_metric(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        batcher_mock = mock.Mock()
        meter.batcher = batcher_mock
        counter = metrics.Counter("name", "desc", "unit", float, meter, False)
        labels = {"key1": "value1"}
        meter.register_view(View(counter, SumAggregator))
        counter.add(1.0, labels)
        meter.collect()
        self.assertFalse(batcher_mock.process.called)

    def test_collect_observers(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        batcher_mock = mock.Mock()
        meter.batcher = batcher_mock

        def callback(observer):
            self.assertIsInstance(observer, metrics_api.Observer)
            observer.observe(45, {})

        observer = metrics.ValueObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        meter.observers.add(observer)
        meter.collect()
        self.assertTrue(batcher_mock.process.called)

    def test_record_batch(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        labels = {"key1": "value1", "key2": "value2", "key3": "value3"}
        counter = metrics.Counter("name", "desc", "unit", float, meter)
        valuerecorder = metrics.ValueRecorder(
            "name", "desc", "unit", float, meter
        )
        counter_v = View(counter, SumAggregator)
        measure_v = View(valuerecorder, MinMaxSumCountAggregator)
        meter.register_view(counter_v)
        meter.register_view(measure_v)
        record_tuples = [(counter, 1.0), (valuerecorder, 3.0)]
        meter.record_batch(labels, record_tuples)
        labels_key = metrics.get_dict_as_key(labels)
        self.assertEqual(
            counter.bound_instruments[labels_key]
            .view_datas.pop()
            .aggregator.current,
            1.0,
        )
        self.assertEqual(
            valuerecorder.bound_instruments[labels_key]
            .view_datas.pop()
            .aggregator.current,
            (3.0, 3.0, 3.0, 1),
        )

    def test_create_metric(self):
        resource = mock.Mock(spec=resources.Resource)
        meter_provider = metrics.MeterProvider(resource=resource)
        meter = meter_provider.get_meter(__name__)
        counter = meter.create_metric(
            "name", "desc", "unit", int, metrics.Counter
        )
        self.assertIsInstance(counter, metrics.Counter)
        self.assertEqual(counter.value_type, int)
        self.assertEqual(counter.name, "name")
        self.assertIs(counter.meter.resource, resource)
        self.assertEqual(counter.meter, meter)

    def test_create_updowncounter(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        updowncounter = meter.create_metric(
            "name", "desc", "unit", float, metrics.UpDownCounter, ()
        )
        self.assertIsInstance(updowncounter, metrics.UpDownCounter)
        self.assertEqual(updowncounter.value_type, float)
        self.assertEqual(updowncounter.name, "name")

    def test_create_valuerecorder(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        valuerecorder = meter.create_metric(
            "name", "desc", "unit", float, metrics.ValueRecorder
        )
        self.assertIsInstance(valuerecorder, metrics.ValueRecorder)
        self.assertEqual(valuerecorder.value_type, float)
        self.assertEqual(valuerecorder.name, "name")
        self.assertEqual(valuerecorder.meter, meter)

    def test_register_observer(self):
        meter = metrics.MeterProvider().get_meter(__name__)

        callback = mock.Mock()

        observer = meter.register_observer(
            callback, "name", "desc", "unit", int, metrics.ValueObserver
        )

        self.assertIsInstance(observer, metrics_api.Observer)
        self.assertEqual(len(meter.observers), 1)

        self.assertEqual(observer.callback, callback)
        self.assertEqual(observer.name, "name")
        self.assertEqual(observer.description, "desc")
        self.assertEqual(observer.unit, "unit")
        self.assertEqual(observer.value_type, int)
        self.assertEqual(observer.label_keys, ())
        self.assertTrue(observer.enabled)

    def test_unregister_observer(self):
        meter = metrics.MeterProvider().get_meter(__name__)

        callback = mock.Mock()

        observer = meter.register_observer(
            callback, "name", "desc", "unit", int, metrics.ValueObserver
        )

        meter.unregister_observer(observer)
        self.assertEqual(len(meter.observers), 0)


class TestMetric(unittest.TestCase):
    def test_bind(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric_types = [metrics.Counter, metrics.ValueRecorder]
        labels = {"key": "value"}
        key_labels = metrics.get_dict_as_key(labels)
        for _type in metric_types:
            metric = _type("name", "desc", "unit", int, meter)
            bound_instrument = metric.bind(labels)
            self.assertEqual(
                metric.bound_instruments.get(key_labels), bound_instrument
            )


class TestCounter(unittest.TestCase):
    def test_add(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric = metrics.Counter("name", "desc", "unit", int, meter)
        labels = {"key": "value"}
        counter_v = View(metric, SumAggregator)
        meter.register_view(counter_v)
        bound_mock = metric.bind(labels)
        metric.add(3, labels)
        metric.add(2, labels)
        self.assertEqual(bound_mock.view_datas.pop().aggregator.current, 5)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_add_non_decreasing_int_error(self, logger_mock):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric = metrics.Counter("name", "desc", "unit", int, meter)
        labels = {"key": "value"}
        counter_v = View(metric, SumAggregator)
        meter.register_view(counter_v)
        bound_counter = metric.bind(labels)
        metric.add(3, labels)
        metric.add(0, labels)
        metric.add(-1, labels)
        self.assertEqual(bound_counter.view_datas.pop().aggregator.current, 3)
        self.assertEqual(logger_mock.warning.call_count, 1)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_add_non_decreasing_float_error(self, logger_mock):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric = metrics.Counter("name", "desc", "unit", float, meter)
        labels = {"key": "value"}
        counter_v = View(metric, SumAggregator)
        meter.register_view(counter_v)
        bound_counter = metric.bind(labels)
        metric.add(3.3, labels)
        metric.add(0.0, labels)
        metric.add(0.1, labels)
        metric.add(-0.1, labels)
        self.assertEqual(
            bound_counter.view_datas.pop().aggregator.current, 3.4
        )
        self.assertEqual(logger_mock.warning.call_count, 1)


class TestUpDownCounter(unittest.TestCase):
    def test_add(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric = metrics.UpDownCounter("name", "desc", "unit", int, meter)
        labels = {"key": "value"}
        bound_counter = metric.bind(labels)
        counter_v = View(metric, SumAggregator)
        meter.register_view(counter_v)
        metric.add(3, labels)
        metric.add(2, labels)
        self.assertEqual(bound_counter.view_datas.pop().aggregator.current, 5)


class TestValueRecorder(unittest.TestCase):
    def test_record(self):
        meter = metrics.MeterProvider().get_meter(__name__)
        metric = metrics.ValueRecorder("name", "desc", "unit", int, meter)
        labels = {"key": "value"}
        measure_v = View(metric, MinMaxSumCountAggregator)
        bound_valuerecorder = metric.bind(labels)
        meter.register_view(measure_v)
        values = (37, 42, 7)
        for val in values:
            metric.record(val, labels)
        self.assertEqual(
            bound_valuerecorder.view_datas.pop().aggregator.current,
            (min(values), max(values), sum(values), len(values)),
        )


class TestSumObserver(unittest.TestCase):
    def test_observe(self):
        observer = metrics.SumObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        key_labels = metrics.get_dict_as_key(labels)
        values = (37, 42, 60, 100)
        for val in values:
            observer.observe(val, labels)

        self.assertEqual(observer.aggregators[key_labels].current, values[-1])

    def test_observe_disabled(self):
        observer = metrics.SumObserver(
            None, "name", "desc", "unit", int, ("key",), False
        )
        labels = {"key": "value"}
        observer.observe(37, labels)
        self.assertEqual(len(observer.aggregators), 0)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_observe_incorrect_type(self, logger_mock):
        observer = metrics.SumObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        observer.observe(37.0, labels)
        self.assertEqual(len(observer.aggregators), 0)
        self.assertTrue(logger_mock.warning.called)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_observe_non_decreasing_error(self, logger_mock):
        observer = metrics.SumObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        observer.observe(37, labels)
        observer.observe(14, labels)
        self.assertEqual(len(observer.aggregators), 1)
        self.assertTrue(logger_mock.warning.called)

    def test_run(self):
        callback = mock.Mock()
        observer = metrics.SumObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertTrue(observer.run())
        callback.assert_called_once_with(observer)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_run_exception(self, logger_mock):
        callback = mock.Mock()
        callback.side_effect = Exception("We have a problem!")

        observer = metrics.SumObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertFalse(observer.run())
        self.assertTrue(logger_mock.warning.called)


class TestUpDownSumObserver(unittest.TestCase):
    def test_observe(self):
        observer = metrics.UpDownSumObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        key_labels = metrics.get_dict_as_key(labels)
        values = (37, 42, 14, 30)
        for val in values:
            observer.observe(val, labels)

        self.assertEqual(observer.aggregators[key_labels].current, values[-1])

    def test_observe_disabled(self):
        observer = metrics.UpDownSumObserver(
            None, "name", "desc", "unit", int, ("key",), False
        )
        labels = {"key": "value"}
        observer.observe(37, labels)
        self.assertEqual(len(observer.aggregators), 0)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_observe_incorrect_type(self, logger_mock):
        observer = metrics.UpDownSumObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        observer.observe(37.0, labels)
        self.assertEqual(len(observer.aggregators), 0)
        self.assertTrue(logger_mock.warning.called)

    def test_run(self):
        callback = mock.Mock()
        observer = metrics.UpDownSumObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertTrue(observer.run())
        callback.assert_called_once_with(observer)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_run_exception(self, logger_mock):
        callback = mock.Mock()
        callback.side_effect = Exception("We have a problem!")

        observer = metrics.UpDownSumObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertFalse(observer.run())
        self.assertTrue(logger_mock.warning.called)


class TestValueObserver(unittest.TestCase):
    def test_observe(self):
        observer = metrics.ValueObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        key_labels = metrics.get_dict_as_key(labels)
        values = (37, 42, 7, 21)
        for val in values:
            observer.observe(val, labels)
        self.assertEqual(
            observer.aggregators[key_labels].mmsc.current,
            (min(values), max(values), sum(values), len(values)),
        )

        self.assertEqual(observer.aggregators[key_labels].current, values[-1])

    def test_observe_disabled(self):
        observer = metrics.ValueObserver(
            None, "name", "desc", "unit", int, ("key",), False
        )
        labels = {"key": "value"}
        observer.observe(37, labels)
        self.assertEqual(len(observer.aggregators), 0)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_observe_incorrect_type(self, logger_mock):
        observer = metrics.ValueObserver(
            None, "name", "desc", "unit", int, ("key",), True
        )
        labels = {"key": "value"}
        observer.observe(37.0, labels)
        self.assertEqual(len(observer.aggregators), 0)
        self.assertTrue(logger_mock.warning.called)

    def test_run(self):
        callback = mock.Mock()
        observer = metrics.ValueObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertTrue(observer.run())
        callback.assert_called_once_with(observer)

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_run_exception(self, logger_mock):
        callback = mock.Mock()
        callback.side_effect = Exception("We have a problem!")

        observer = metrics.ValueObserver(
            callback, "name", "desc", "unit", int, (), True
        )

        self.assertFalse(observer.run())
        self.assertTrue(logger_mock.warning.called)


# pylint: disable=no-self-use
class TestBoundCounter(unittest.TestCase):
    def test_add(self):
        meter_mock = mock.Mock()
        metric_mock = mock.Mock()
        metric_mock.enabled = True
        metric_mock.value_type = int
        metric_mock.meter = meter_mock
        bound_metric = metrics.BoundCounter((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_metric.view_datas = [view_datas_mock]
        bound_metric.add(3)
        view_datas_mock.record.assert_called_once_with(3)

    def test_add_disabled(self):
        meter_mock = mock.Mock()
        metric_mock = mock.Mock()
        metric_mock.enabled = False
        metric_mock.value_type = int
        metric_mock.meter = meter_mock
        bound_metric = metrics.BoundCounter((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_metric.view_datas = [view_datas_mock]
        bound_metric.add(3)
        view_datas_mock.record.update_view.assert_not_called()

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_add_incorrect_type(self, logger_mock):
        meter_mock = mock.Mock()
        viewm_mock = mock.Mock()
        meter_mock.view_manager = viewm_mock
        metric_mock = mock.Mock()
        metric_mock.enabled = True
        metric_mock.value_type = float
        metric_mock.meter = meter_mock
        bound_metric = metrics.BoundCounter((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_metric.view_datas = [view_datas_mock]
        bound_metric.add(3)
        view_datas_mock.record.update_view.assert_not_called()
        self.assertTrue(logger_mock.warning.called)


class TestBoundValueRecorder(unittest.TestCase):
    def test_record(self):
        meter_mock = mock.Mock()
        metric_mock = mock.Mock()
        metric_mock.enabled = True
        metric_mock.value_type = int
        metric_mock.meter = meter_mock
        bound_valuerecorder = metrics.BoundValueRecorder((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_valuerecorder.view_datas = [view_datas_mock]
        bound_valuerecorder.record(3)
        view_datas_mock.record.assert_called_once_with(3)

    def test_record_disabled(self):
        meter_mock = mock.Mock()
        metric_mock = mock.Mock()
        metric_mock.enabled = False
        metric_mock.value_type = int
        metric_mock.meter = meter_mock
        bound_valuerecorder = metrics.BoundValueRecorder((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_valuerecorder.view_datas = [view_datas_mock]
        bound_valuerecorder.record(3)
        view_datas_mock.record.update_view.assert_not_called()

    @mock.patch("opentelemetry.sdk.metrics.logger")
    def test_record_incorrect_type(self, logger_mock):
        meter_mock = mock.Mock()
        metric_mock = mock.Mock()
        metric_mock.enabled = True
        metric_mock.value_type = float
        metric_mock.meter = meter_mock
        bound_valuerecorder = metrics.BoundValueRecorder((), metric_mock)
        view_datas_mock = mock.Mock()
        bound_valuerecorder.view_datas = [view_datas_mock]
        bound_valuerecorder.record(3)
        view_datas_mock.record.update_view.assert_not_called()
        self.assertTrue(logger_mock.warning.called)
